/*
 * Copyright (c) 2016-2017 Illumon and Patent Pending
 */

#include "TradeWriter.hpp"

/*
 * On construction, the TradeWriter declares the columns that are part of the
 * log file.  These must match the schema definition.  In this case, the schema
 * used is:
 *
 *  <Table name="TradeNyseStock" namespace="MarketUs" storageType="NestedPartitionedOnDisk">
 *    <Partitions keyFormula="__PARTITION_AUTOBALANCE_SINGLE__"/>
 *    <Column name="Date" dataType="String" columnType="Partitioning"/>
 *    <Column name="Timestamp" dataType="DateTime" columnType="Normal"/>
 *    <Column name="Exchange" dataType="String" columnType="Normal"/>
 *    <Column name="USym" dataType="String" columnType="Grouping"/>
 *    <Column name="Sym" dataType="String" columnType="Grouping"/>
 *    <Column name="Last" dataType="double" columnType="Normal"/>
 *    <Column name="Size" dataType="int" columnType="Normal"/>
 *    <Column name="ExchangeTimestamp" dataType="DateTime" columnType="Normal"/>
 *  </Table>
 *
 * The "Date" column is not included, because it is a Partitioning column, and
 * the data import server implicitly stores the partitioning column as part of
 * the directory structure.  Similarly, values with schema attribute of
 * intradayType="none" should not be included.
 *
 * All other columns must be included with the same type (Boolean, Byte, Char,
 * Double, Float, Int, Long, String) as in the schema definition.
 *
 * DateTime columns are represented as Long nano seconds since the unix epoch.
 *
 * Blobs and Enums are not supported from C++.
 */
TradeWriter::TradeWriter() {
  addColumn("Timestamp", illumon::binarystore::SupportedType::Long);
  addColumn("Exchange", illumon::binarystore::SupportedType::String);
  addColumn("USym", illumon::binarystore::SupportedType::String);
  addColumn("Sym", illumon::binarystore::SupportedType::String);
  addColumn("Side", illumon::binarystore::SupportedType::String);
  addColumn("Last", illumon::binarystore::SupportedType::Double);
  addColumn("Size", illumon::binarystore::SupportedType::Int);
  addColumn("ExchangeTimestamp", illumon::binarystore::SupportedType::Long);
}

/*
 * This is a thin wrapper to the underlying writeRow in BinaryStoreWriter.
 *
 * The arguments to writeRow must be in the same order that the constructor
 * declared the columns using addColumn; otherwise a logic_error will be
 * thrown at runtime.
 */
void TradeWriter::writeTrade(uint64_t insertTime, std::string const & usym, std::string const & sym, std::string const & side, double price, int size, uint64_t exchangeTimestamp, std::string const & exchange)
{
  writeRow(
      insertTime,
      exchange,
      usym,
      sym,
      side,
      price,
      size,
      exchangeTimestamp);
}
