/*
 * Copyright (c) 2016-2017 Illumon and Patent Pending
 */

#include "TradeWriter.hpp"

#include <fstream>
#include <iostream>
#include <map>
#include <stdexcept>
#include <vector>

/*
 * Utility method to read a CSV into a vector of vectors.
 */
std::pair<std::vector<std::vector<std::string>>, std::map<std::string, int>> readCsv(std::ifstream & inFile) {
  std::map<std::string, int> headerMap;
  std::map<int, std::string> reverseHeaderMap;
  std::vector<std::vector<std::string>> result;

  std::string line;

  bool gotHeader = static_cast<bool>(std::getline(inFile, line));
  if (!gotHeader) {
    throw std::runtime_error("No header line found!");
  }

  std::istringstream iss(line);

  std::string token;
  int fieldNum = 0;
  while (std::getline(iss, token, ',')) {
    headerMap[token] = fieldNum;
    reverseHeaderMap[fieldNum++] = token;
  }

  while (std::getline(inFile, line)) {
    if (line.empty()) {
      continue;
    }

    iss.str(line);
    iss.clear();

    std::vector<std::string> row;
    row.reserve(headerMap.size());
    while (std::getline(iss, token, ',')) {
      if (!token.empty() && token[0] == '"' && token[token.size() - 1] == '"') {
        token.resize(token.size() - 1);
        token.erase(token.begin(), token.begin() + 1);
      }
      row.push_back(token);
    }

    if (row.size() != headerMap.size()) {
      std::ostringstream oss;
      oss << "Row has " << row.size() << " fields, but " << headerMap.size() << " are required: " << line;
      throw std::runtime_error(oss.str());
    }

    result.push_back(row);
  }

  return std::make_pair(result, headerMap);
}

/* Utility method to convert a string to an unsigned long. */
uint64_t convertLong(std::string const & stringValue, std::string const & name) {
    char *endptr;
    uint64_t value = strtoull(stringValue.c_str(), &endptr, 10);
    if (*endptr) {
      throw std::runtime_error("Can not parse " + name + ": " + stringValue);
    }
    return value;
}

/* Utility method to convert a string to an double long. */
double convertDouble(std::string const & stringValue, std::string const & name) {
    char *endptr;
    double value = strtod(stringValue.c_str(), &endptr);
    if (*endptr) {
      throw std::runtime_error("Can not parse " + name + ": " + stringValue);
    }
    return value;
}

int main(int argc, char *argv[]) {
  std::string inFileName("trades.csv");
  std::string outFileName("trades-out.bin");
  std::string origFileName("trades-in.bin");

  // open a CSV file that we'll translate into a bin file
  std::ifstream inFile(inFileName);

  std::pair<std::vector<std::vector<std::string>>, std::map<std::string, int>> result;

  try {
    result = readCsv(inFile);
  } catch (std::exception & ex) {
    std::cerr << "Could not read CSV " << inFileName << ": " << ex.what() << std::endl;
    exit(1);
  }

  // we've got our CSV loaded, make some const copies of the map and vector so
  // we don't accidentally modify it below
  auto const & headerMap = result.second;
  auto const & rows = result.first;

  // create a trade writer, which initializes the fields for us in the constructor
  TradeWriter tradewriter;

  try {
    // open the output file, which writes the header
    tradewriter.open(outFileName);
  } catch (std::exception & ex) {
    std::cerr << "Could not open output bin: " << ex.what() << std::endl;
    exit(1);
  }

  // loop through each row, pull out the fields and convert them to the appropriate type
  for (auto & row : rows) {
    if (row.size() != headerMap.size()) {
      throw std::logic_error("Invalid row size!");
    }

    try {
      // times are stored as a long nanos since the epoch
      uint64_t insertTime = convertLong(row.at(headerMap.at("Timestamp")), "Timestamp");
      std::string const & usym = row.at(headerMap.at("USym"));
      std::string const & sym = row.at(headerMap.at("Sym"));
      std::string const & side = row.at(headerMap.at("Side"));
      double price = convertDouble(row.at(headerMap.at("Last")), "Last");
      int size = convertLong(row.at(headerMap.at("Size")), "Size");
      uint64_t exchangeTimestamp = convertLong(row.at(headerMap.at("ExchangeTimestamp")), "ExchangeTimestamp");
      std::string const & exchange = row.at(headerMap.at("Exchange"));

      tradewriter.writeTrade(insertTime, usym, sym, side, price, size, exchangeTimestamp, exchange);
    } catch (std::exception & ex) {
      std::cerr << "Error converting row: " << ex.what() << std::endl;
      exit(1);
    }
  }

  // close the output file, releasing it's resources
  tradewriter.close();

  // now open the output file using regular I/O so we can compare it to the
  // original .bin file; verifying that our library is producing identical data
  // as the original

  std::ifstream origFile(origFileName, std::ifstream::binary);
  std::ifstream newFile(outFileName, std::ifstream::binary);

  origFile.seekg(0, origFile.end);
  off_t origLen = origFile.tellg();
  origFile.seekg(0, origFile.beg);

  newFile.seekg(0, newFile.end);
  off_t newLen = newFile.tellg();
  newFile.seekg(0, newFile.beg);

  if (origLen != newLen) {
    std::cerr << "New file length: " << newLen << " not equal to orginal length " << origLen;
    exit(1);
  }


  std::vector<char> origContent;
  std::vector<char> newContent;

  origContent.resize(origLen);
  newContent.resize(newLen);

  origFile.read(origContent.data(), origLen);
  newFile.read(newContent.data(), newLen);

  if (origContent != newContent) {
    std::cerr << "File contents differ!";
    exit(1);
  } else {
    std::cout << "Contents are identical (" << rows.size() << " rows, " << origLen << " bytes)." << std::endl;
  }

  origFile.close();
  newFile.close();

  exit(0);
}
