/*
 * Copyright (c) 2016-2017 Illumon and Patent Pending
 */
 
#pragma once

// BinaryStoreWriter is a C++ implementation of the Java BinaryStoreWriter used
// by the IntradayLoggerFactor generated classes.
//
// The code is somewhat structured like the Java version in that this class
// doesn't know anything about the data that you are trying to log to a
// specific file.  Instead, you can extend it to provide a friendlier
// interface.
//
// In your derived constructor, you should call addColumn with each column that
// you would like to write to the file.  Then when you call writeRow, which is
// a templated-function, it expects the arguments in the same order.  The writeRow
// function will throw an exception if the type of your argument does not match
// the type that is expected at that position.  To avoid this, one possibility
// is is to have a more specific function in your derived class that will
// translate an input structure to the proper order of fields.
//
// Each row results in a write(2) system call, but no synchronization to disk
// is performed.  Your first row and the header will also result in
// allocations.  If you increase the total size of a row's serialized
// representation, an allocation is also required.
//
// Only primitive types and strings are supported.
//
// Blobs, Enums, and Null String values are not supported.  Null primitives can
// be written using the NULL_CHAR, NULL_BYTE, NULL_SHORT, NULL_INT, NULL_FLOAT,
// and NULL_DOUBLE constants.

#include <ctime>
#include <cstring>
#include <stdexcept>
#include <sstream>
#include <string>
#include <vector>
#include <limits>

#include <sys/fcntl.h>
#include <sys/types.h>

#include <boost/optional.hpp>

#include "likely.hpp"
#include "FileDescriptor.hpp"
#include "CountArgs.hpp"

namespace illumon {
namespace binarystore {

  static const uint8_t NULL_CHAR = 0;
  static const int8_t NULL_BYTE = std::numeric_limits<int8_t>::min(); // -128
  static const int16_t NULL_SHORT = std::numeric_limits<int16_t>::min(); // -32768
  static const int32_t NULL_INT = std::numeric_limits<int32_t>::min(); // -2147483648
  static const int64_t NULL_LONG = std::numeric_limits<int64_t>::min();  // -9223372036854775808
  static const float NULL_FLOAT = (-std::numeric_limits<float>::max()); // -3.40282e+38
  static const double NULL_DOUBLE = (-std::numeric_limits<double>::max()); //-1.79769e+308

// An enumeration which contains each type that may be written to the database.
// In practice, not all of these are implemented, but all of them are defined
// in Java.
//
// Strings, Integers, and Doubles are supported.
enum class SupportedType
{
  Boolean,
  Byte,
  Char,
  Double,
  Float,
  Int,
  Long,
  Short,
  Blob,
  String,
  Enum,
};

// Forward declaration of the BinaryStoreWriter
class BinaryStoreWriter;

namespace detail {
    // You must specialize the putValue function for each type of argument you want to use.
    //
    // the second int argument is the index of the value we are putting (used for nulling out the presence bitmap)
    //
    // The SupportedType parameter is used to verify that the type of this value to write matches the header.
    template <typename T> void putValue(BinaryStoreWriter &, int, SupportedType, T);
    template <> void putValue<int8_t>(BinaryStoreWriter &, int, SupportedType, int8_t);
    template <> void putValue<uint8_t>(BinaryStoreWriter &, int, SupportedType, uint8_t);
    template <> void putValue<int16_t>(BinaryStoreWriter &, int, SupportedType, int16_t);
    template <> void putValue<uint16_t>(BinaryStoreWriter &, int, SupportedType, uint16_t);
    template <> void putValue<int32_t>(BinaryStoreWriter &, int, SupportedType, int32_t);
    template <> void putValue<uint32_t>(BinaryStoreWriter &, int, SupportedType, uint32_t);
    template <> void putValue<int64_t>(BinaryStoreWriter &, int, SupportedType, int64_t);
    template <> void putValue<uint64_t>(BinaryStoreWriter &, int, SupportedType, uint64_t);
    template <> void putValue<char const *>(BinaryStoreWriter &, int, SupportedType, char const *);
    template <> void putValue<char *>(BinaryStoreWriter &, int, SupportedType, char *);
    template <> void putValue<std::string const &>(BinaryStoreWriter &, int, SupportedType, std::string const &);
    template <> void putValue<boost::optional<std::string const &>>(BinaryStoreWriter & bsw, int index, SupportedType type, boost::optional<std::string const &>);
    template <> void putValue<boost::optional<std::string>>(BinaryStoreWriter & bsw, int index, SupportedType type, boost::optional<std::string>);
    template <> void putValue<boost::none_t>(BinaryStoreWriter & bsw, int index, SupportedType type, boost::none_t);
    template <> void putValue<double>(BinaryStoreWriter &, int, SupportedType, double);
    template <> void putValue<float>(BinaryStoreWriter &, int, SupportedType, float);
    template <> void putValue<bool>(BinaryStoreWriter &, int, SupportedType, bool);
    #ifdef __APPLE__
    template <> void putValue<long>(BinaryStoreWriter &, int, SupportedType, long);
    #else
    template <> void putValue<long long>(BinaryStoreWriter &, int, SupportedType, long long);
    #endif
    template <> void putValue<char>(BinaryStoreWriter &, int, SupportedType, char);

    void nullValue(BinaryStoreWriter &, int);

    // anything that we prefer to pass as a reference, should be defined
    template <typename T> struct type_to_const_ref
    {
      typedef T value_type;
    };

    template <> struct type_to_const_ref<std::string>
    {
      typedef std::string const & value_type;
    };

    template <> struct type_to_const_ref<std::string &>
    {
      typedef std::string const & value_type;
    };

    template <typename T> struct optional_to_type
    {
      typedef T value_type;
    };

    template <typename T> struct optional_to_type<boost::optional<T>>
    {
      typedef T value_type;
    };

    template <typename T> struct Putter {
      void operator()(BinaryStoreWriter & bsw, int index, SupportedType type, T const & t) const {
        putValue<typename type_to_const_ref<T>::value_type>(bsw, index, type, t);
      }
      void operator()(BinaryStoreWriter & bsw, int index, SupportedType type, boost::optional<T> const & t) const {
        if (t) {
          operator()(bsw, index, type, *t);
        } else {
          nullValue(bsw, index);
        }
      }
    };

}

class BinaryStoreWriter
{
  public:
    // open filename, write the header, and be prepared for subsequent calls to writeRow,
    // you must call addColumn first
    void open(std::string const & filename);

    //Open the file for logging, rotating log files at the specified interval
    void open(std::string const & filename, time_t interval);

    // close the binary store
    void close();

    bool isOpen() {
      return m_fd.isOpen();
    }

    // add a column to this writer, the columns are added in order of the calls.
    // this must be called before open
    void addColumn(std::string const & name, SupportedType type);

    // write a row to the file, the arguments are the fields, in the same order as calls to addColumn;
    template <typename... Values> void writeRow(Values... args);

  protected:
    virtual bool shouldRotate();
    virtual std::string getNextFileName();

  private:
    // Description of what we're writing to this file
    struct ColInfo
    {
        ColInfo(std::string a_name, SupportedType a_type)
          : name(a_name)
          , type(a_type)
        {
        }

        std::string name;
        SupportedType type;
    };

    std::vector <ColInfo> m_columns;
    std::string fileName;

    // The open file that we are writing to
    util::FileDescriptor m_fd;

    // A buffer of bytes that we put the row that we are currently building up into.
    std::vector<char> m_byteBuffer;

    time_t rotationPeriod;
    time_t curInterval;

    // peel off one column and write it out
    template <typename T, typename... Values>
    void internalWriteRow(int index, T const & arg1, Values... rest)
    {
      detail::Putter<typename detail::optional_to_type<T>::value_type>()(*this, index, m_columns[index].type, arg1);
      internalWriteRow(++index, rest...);
    }

    // base case where we have written all of the columns
    void internalWriteRow(int index)
    {
    }

    // enqueue some data for writing to the file
    void bufferBytes(void const *bytes, size_t size);

    // flush anything buffered to the file, may not actually sync to disk
    void flushWrite();

    //Close the current log and open a new one
    void rotateFile();

    // cause a hard failure if there is anything left in the byte buffer when we are beginning a new row.
    void checkBufferEmpty() const;

    // Append the specified primitive type to the buffer.
    void putByte(uint8_t value);
    void putInteger(uint32_t value);
    void putLong(uint64_t value);
    void putShort(uint16_t value);
    void putFloat(float value);
    void putDouble(double value);
    void putChar(char value);
    void putBool(bool value);

    // null out a value in the presence map
    void nullValue(size_t index);

    template <class T> friend void detail::putValue(BinaryStoreWriter & bsw, int, SupportedType type, T);
    friend void detail::nullValue(BinaryStoreWriter & bsw, int);
};

std::ostream & operator<<(std::ostream & os, SupportedType type);

// given a SupportedType input, return the size and string that should be written to the header
size_t getTypeSize(SupportedType type);
char const *getTypeString(SupportedType type);


template <typename... Values> void BinaryStoreWriter::writeRow(Values... args)
{
  checkBufferEmpty();

  if (unlikely(!m_fd.isOpen()))
  {
    throw std::logic_error("Can not call writeRow on an unopened BinaryStoreWriter!");
  }

  size_t const fieldCount = CountArgs<Values...>::value;

  if (unlikely(fieldCount != m_columns.size()))
  {
    std::ostringstream oss;
    oss << "Incorrect number of arguments passed to BinaryStoreWriter: was "
        << fieldCount << ", expected " << m_columns.size();
    throw std::logic_error(oss.str());
  }

  // the header for each row is a mask that contains whether the following fields actually exist
  for (size_t ii = 0; ii < fieldCount; ii += 8)
  {
    uint8_t byte = (ii + 8 <= fieldCount) ? 0xff : (1 << (fieldCount % 8)) - 1;
    putByte(byte);
  }

  internalWriteRow(0, args...);

  if(shouldRotate()) {
    rotateFile();
  }

  // do the actual write system call.
  flushWrite();
}

}} // namespace illumon::binarystore
