/*
 * Copyright (c) 2016-2017 Illumon and Patent Pending
 */

#include "BinaryStoreWriter.hpp"

#include "ErrnoException.hpp"

#include <cassert>
#ifdef __APPLE__
#include <libkern/OSByteOrder.h>
#else
#include <endian.h>
#include <byteswap.h>
#endif

#include <unistd.h>

// if endian.h has not created the conversion functions for us, we need to do
// it ourselves to support CentOS 5.4.  On CentOS 6.2 endian.h provides these
// functions for us already.  This trickery depends on htobe64 being a macro,
// which it just happens to be.

#ifndef htobe64

#if __BYTE_ORDER == __LITTLE_ENDIAN
#ifdef __APPLE__
#define htobe16(value) OSSwapInt16(value)
#define htobe32(value) OSSwapInt32(value)
#define htobe64(value) OSSwapInt64(value)
#else
#define htobe16(value) __bswap_16(value)
#define htobe32(value) __bswap_32(value)
#define htobe64(value) __bswap_64(value)
#endif
#else

#define htobe16(value) (value)
#define htobe32(value) (value)
#define htobe64(value) (value)

#endif

#endif // htobe64 !defined


namespace illumon {
namespace binarystore {

std::ostream & operator<<(std::ostream & os, SupportedType type)
{
  return os << getTypeString(type);
}

void BinaryStoreWriter::addColumn(std::string const & name, SupportedType type)
{
  if (m_fd.isOpen())
  {
    throw std::logic_error("Can not call addColumn after openFile on BinaryStoreWriter!");
  }

  switch (type)
  {
    case SupportedType::Enum:
    case SupportedType::Blob:
      {
        std::ostringstream oss;
        oss << type << " is not supported by BinaryStoreWriter.";
        throw std::runtime_error(oss.str());
      }
      break;
    default:
      // we handle the primitive types
      break;
  }

  m_columns.push_back(ColInfo(name, type));
}

void BinaryStoreWriter::open(std::string const & filename) {
  open(filename,0);
}

void BinaryStoreWriter::open(std::string const & filename, time_t interval)
{
  if (m_columns.empty())
  {
    throw std::logic_error("can not call openFile before defining columns!");
  }

  if (m_fd.isOpen())
  {
    throw std::logic_error("Can not call openFile twice on a BinaryStoreWriter!");
  }

  this->fileName = filename;

  rotationPeriod = interval;

  rotateFile();
}

void BinaryStoreWriter::rotateFile() {
  std::string useName = getNextFileName();

  if(m_fd.isOpen()) {
    flushWrite();
  }

  int rawFd = ::open(useName.c_str(), O_RDWR|O_CREAT|O_EXCL, 0644);
  if (rawFd < 0)
  {
    throw util::errno_exception("Can not open " + useName, errno);
  }
  m_fd.reset(rawFd);

  checkBufferEmpty();

  putInteger(m_columns.size());

  for (ColInfo ci : m_columns)
  {
    // write out the name, including the \0
    bufferBytes(ci.name.c_str(), ci.name.size() + 1);
    // write out the type, including the \0
    char const *typeString = getTypeString(ci.type);
    int len = strlen(typeString);
    bufferBytes(typeString, len + 1);

    // write out the size of the type
    putInteger(getTypeSize(ci.type));

    // if we supported enum types, the meta data would go here, in its absence
    // we write a size of 0
    putInteger(0);
  }

  flushWrite();
}

void BinaryStoreWriter::close()
{
  checkBufferEmpty();
  m_fd.reset();
}

size_t getTypeSize(SupportedType type)
{
  switch (type)
  {
    case SupportedType::Boolean: return 1;
    case SupportedType::Byte: return 1;
    case SupportedType::Char: return 1;
    case SupportedType::Double: return 8;
    case SupportedType::Float: return 4;
    case SupportedType::Int: return 4;
    case SupportedType::Long: return 8;
    case SupportedType::Short: return 2;
    case SupportedType::Blob: return 0;
    case SupportedType::String: return 0;
    case SupportedType::Enum: return 4;
  }
  return 0;
}

char const *getTypeString(SupportedType type)
{
  switch (type)
  {
    case SupportedType::Boolean: return "Boolean";
    case SupportedType::Byte: return "Byte";
    case SupportedType::Char: return "Char";
    case SupportedType::Double: return "Double";
    case SupportedType::Float: return "Float";
    case SupportedType::Int: return "Int";
    case SupportedType::Long: return "Long";
    case SupportedType::Short: return "Short";
    case SupportedType::Blob: return "Blob";
    case SupportedType::String: return "String";
    case SupportedType::Enum: return "Enum";
  }
  return "";
}

void BinaryStoreWriter::flushWrite()
{
  size_t written = 0;

  while (written < m_byteBuffer.size())
  {
    ssize_t result = write(m_fd, m_byteBuffer.data() + written, m_byteBuffer.size() - written);
    if (result < 0)
    {
      throw util::errno_exception("Write to BinaryStoreWriter failed(): ", errno);
    }
    if (result == 0)
    {
      throw std::runtime_error("Write to BinaryStoreWriter did not write any data.");
    }
    written += result;
  }

  m_byteBuffer.clear();
}

void BinaryStoreWriter::checkBufferEmpty() const
{
  assert(m_byteBuffer.empty());
}

void BinaryStoreWriter::bufferBytes(void const *bytes, size_t size)
{
  std::copy(static_cast<char const *>(bytes), static_cast<char const *>(bytes) + size, std::back_inserter(m_byteBuffer));
}

bool BinaryStoreWriter::shouldRotate() {
  if(rotationPeriod != 0) {
    std::time_t now = std::time(nullptr);
    return (now >= curInterval + rotationPeriod);
  }

  return false;
}

std::string BinaryStoreWriter::getNextFileName() {
  if(rotationPeriod != 0) {
    std::time_t now = std::time(nullptr);
    curInterval = now - (now % rotationPeriod);
    std::tm tmTime = m_fd.isOpen() ? *std::localtime(&curInterval)
                                   : *std::localtime(&now);

    std::ostringstream annotatedFile;
    char tmp[100];
    std::strftime(tmp,sizeof(tmp),"%F-%H%M%S%z", &tmTime);
    annotatedFile<<fileName<<"."<<tmp;

    return annotatedFile.str();
  }
  else return fileName;
}

// nullOut the given index in our presence map
void BinaryStoreWriter::nullValue(size_t index)
{
  if (m_columns.size() <= index) {
    throw std::logic_error("Invalid index.");
  }

  int byteIndex = index / 8;
  int position = index % 8;

  if (m_byteBuffer.size() <= byteIndex) {
    throw std::logic_error("Buffer too short for index.");
  }

  m_byteBuffer[byteIndex] &= static_cast<uint8_t>(~(1<<position));
}

// Append the specified primitive type to the buffer.
void BinaryStoreWriter::putByte(uint8_t value)
{
  bufferBytes(&value, sizeof(value));
}

void BinaryStoreWriter::putInteger(uint32_t value)
{
  value = htobe32(value);
  bufferBytes(&value, sizeof(value));
}

void BinaryStoreWriter::putLong(uint64_t value)
{
  value = htobe64(value);
  bufferBytes(&value, sizeof(value));
}

void BinaryStoreWriter::putShort(uint16_t value)
{
  value = htobe16(value);
  bufferBytes(&value, sizeof(value));
}

void BinaryStoreWriter::putFloat(float value)
{
  uint32_t writeValue = htobe32(*reinterpret_cast<uint32_t const *>(&value));
  static_assert(sizeof(value) == sizeof(writeValue), "Integer/float size mismatch!");
  bufferBytes(&writeValue, sizeof(writeValue));
}

void BinaryStoreWriter::putDouble(double value)
{
  uint64_t writeValue = htobe64(*reinterpret_cast<uint64_t const *>(&value));
  static_assert(sizeof(value) == sizeof(writeValue), "Integer/float size mismatch!");
  bufferBytes(&writeValue, sizeof(writeValue));
}

void BinaryStoreWriter::putChar(char value) {
  uint8_t writeVal = static_cast<uint8_t>(value);
  bufferBytes(&writeVal,sizeof(writeVal));
}

void BinaryStoreWriter::putBool(bool value) {
  uint8_t writeVal = static_cast<uint8_t>(value);
  bufferBytes(&writeVal,sizeof(writeVal));
}

// Forward declare some of the templates
namespace detail {
// Functions that actually write data out to the file
template <> void putValue<int8_t>(BinaryStoreWriter & bsw, int index, SupportedType type, int8_t value)
{
  if (unlikely(type != SupportedType::Byte))
  {
    std::stringstream ostr(std::stringstream::in | std::stringstream::out);
    ostr << "This should not be an Byte. Value " << value;
    throw std::logic_error(ostr.str());
  }

  bsw.putByte(value);
}

template <> void putValue<uint8_t>(BinaryStoreWriter & bsw, int index, SupportedType type, uint8_t value)
{
  if (unlikely(type != SupportedType::Byte))
  {
    std::stringstream ostr(std::stringstream::in | std::stringstream::out);
    ostr << "This should not be a Byte. Value " << value;
    throw std::logic_error(ostr.str());
  }

  bsw.putByte(value);
}

template <> void putValue<int16_t>(BinaryStoreWriter & bsw, int index, SupportedType type, int16_t value)
{
  if (unlikely(type != SupportedType::Short))
  {
    std::stringstream ostr(std::stringstream::in | std::stringstream::out);
    ostr << "This should not be an Short. Value " << value;
    throw std::logic_error(ostr.str());
  }

  bsw.putShort(value);
}

template <> void putValue<uint16_t>(BinaryStoreWriter & bsw, int index, SupportedType type, uint16_t value)
{
  if (unlikely(type != SupportedType::Short))
  {
    std::stringstream ostr(std::stringstream::in | std::stringstream::out);
    ostr << "This should not be a Short. Value " << value;
    throw std::logic_error(ostr.str());
  }

  bsw.putShort(value);
}

template <> void putValue<int32_t>(BinaryStoreWriter & bsw, int index, SupportedType type, int32_t value)
{
  if (unlikely(type != SupportedType::Int))
  {
    std::stringstream ostr(std::stringstream::in | std::stringstream::out);
    ostr << "This should not be an Integer. Value " << value;
    throw std::logic_error(ostr.str());
  }

  bsw.putInteger(value);
}

template <> void putValue<uint32_t>(BinaryStoreWriter & bsw, int index, SupportedType type, uint32_t value)
{
  if (unlikely(type != SupportedType::Int))
  {
    std::stringstream ostr(std::stringstream::in | std::stringstream::out);
    ostr << "This should not be a Int. Value " << value;
    throw std::logic_error(ostr.str());
  }

  bsw.putInteger(value);
}

template <> void putValue<int64_t>(BinaryStoreWriter & bsw, int index, SupportedType type, int64_t value)
{
  if (unlikely(type != SupportedType::Long))
  {
    std::stringstream ostr(std::stringstream::in | std::stringstream::out);
    ostr << "This should not be a Long. Value " << value;
    throw std::logic_error(ostr.str());
  }

  bsw.putLong(value);
}

template <> void putValue<uint64_t>(BinaryStoreWriter & bsw, int index, SupportedType type, uint64_t value)
{
  if (unlikely(type != SupportedType::Long))
  {
    std::stringstream ostr(std::stringstream::in | std::stringstream::out);
    ostr << "This should not be a Long. Value " << value;
    throw std::logic_error(ostr.str());
  }

  bsw.putLong(value);
}

#ifdef __APPLE__
template <> void putValue<long>(BinaryStoreWriter & bsw, int index, SupportedType type, long value)
{
 if (unlikely(type != SupportedType::Long))
 {
   throw std::logic_error("This should not be a Long.");
  }

  bsw.putLong(value);
}
#else
template <> void putValue<long long>(BinaryStoreWriter & bsw, int index, SupportedType type, long long value)
{
 if (unlikely(type != SupportedType::Long))
 {
   throw std::logic_error("This should not be a Long.");
  }

  bsw.putLong(value);
}
#endif

template <> void putValue<char const *>(BinaryStoreWriter & bsw, int index, SupportedType type, char const *string)
{
  if (unlikely(type != SupportedType::String))
  {
    throw std::logic_error("This should not be a string.");
  }

  if (string == nullptr) {
    // we need to null out the value
    bsw.nullValue(index);
    return;
  }

  size_t len = strlen(string);
  bsw.putShort(len);
  bsw.bufferBytes(string, len);
}

template <> void putValue<char *>(BinaryStoreWriter & bsw, int index, SupportedType type, char *string)
{
  putValue(bsw, index, type, static_cast<char const *>(string));
}


template <> void putValue<boost::optional<std::string const &>>(BinaryStoreWriter & bsw, int index, SupportedType type, boost::optional<std::string const &> string)
{
  if (unlikely(type != SupportedType::String))
  {
    throw std::logic_error("This should not be a string.");
  }

  if (string) {
    bsw.putShort(string->size());
    bsw.bufferBytes(string->data(), string->size());
  } else {
    bsw.nullValue(index);
  }
}

template <> void putValue<boost::optional<std::string>>(BinaryStoreWriter & bsw, int index, SupportedType type, boost::optional<std::string> string)
{
  if (unlikely(type != SupportedType::String))
  {
    throw std::logic_error("This should not be a string.");
  }

  if (string) {
    bsw.putShort(string->size());
    bsw.bufferBytes(string->data(), string->size());
  } else {
    bsw.nullValue(index);
  }
}

template <> void putValue<std::string const &>(BinaryStoreWriter & bsw, int index, SupportedType type, std::string const & string)
{
  if (unlikely(type != SupportedType::String))
  {
    throw std::logic_error("This should not be a string.");
  }

  bsw.putShort(string.size());
  bsw.bufferBytes(string.data(), string.size());
}

template <> void putValue<boost::none_t>(BinaryStoreWriter & bsw, int index, SupportedType type, boost::none_t)
{
  bsw.nullValue(index);
}


template <> void putValue<double>(BinaryStoreWriter & bsw, int index, SupportedType type, double value)
{
  if (unlikely(type != SupportedType::Double))
  {
    throw std::logic_error("This should not be a double.");
  }

  bsw.putDouble(value);
}

template <> void putValue<float>(BinaryStoreWriter & bsw, int index, SupportedType type, float value)
{
  if (unlikely(type != SupportedType::Float))
  {
    throw std::logic_error("This should not be a float.");
  }

  bsw.putFloat(value);
}

template <> void putValue<bool>(BinaryStoreWriter & bsw, int index, SupportedType type, bool value)
{
  if (unlikely(type != SupportedType::Boolean))
  {
    throw std::logic_error("This should not be a Bool.");
  }

  bsw.putBool(value);
}

template <>  void putValue<char>(BinaryStoreWriter & bsw, int index, SupportedType type, char value) {
  if(unlikely(type != SupportedType::Char)) {
    throw std::logic_error("This should not be a char");
  }

  bsw.putChar(value);
}

void nullValue(BinaryStoreWriter & bsw, int index) {
  bsw.nullValue(index);
}

}

}} // namespace illumon::binarystore
